import type { WritableComputedRef, InjectionKey, UnwrapNestedRefs } from 'vue'
import { computed, inject, provide, unref } from 'vue'
import type { Block, Category, Module, Page, View, ViewChildren } from './types'
import { isRefer, setValue, valueOf } from './utils'

/** 画布配置 */
export interface CanvasConfig {
  /** 画布是否聚焦 */
  focused: boolean
  /** 画布缩放比例 */
  scale: number
}

export interface PageConfig extends Omit<Page, 'body'> {
}

export type ConfiguratorType =
  | 'page'
  | 'block'
  | 'view'

export interface Exported {
  attrs: PageConfig
  data: Record<string, Record<string, unknown>>
}

export interface EngineContextBase {
  /** 目标 */
  target?: 'app' | 'routine'
  /** 当前选中的模板 */
  activeBlockId?: string
  /** 当前选中的可编辑组件 */
  activeViewId?: string
  /** 鼠标进入的可编辑组件 */
  hoverViewId?: string
  /** 正在被拖拽的模板 */
  draggingViewId?: string
  /** 当前配置 */
  configurator?: ConfiguratorType
  /** 页面配置 */
  page: PageConfig
  /** 画布配置 */
  canvas: CanvasConfig
  /** 可以分类 */
  categories: Category[]
  /** 布局视图列表 */
  blocks: Block[]
  /** 主题列表 */
  views: Record<string, View>
  /** 内部唯一标识 */
  nextId: number
  /** 数据源 */
  sources: Record<string, Record<string, unknown>>
  /** 导出数据 */
  export: () => Exported
  /** 获取动态数据 */
  value: <T>(blockId: string, key: string) => T | undefined
}

export type EngineContext = UnwrapNestedRefs<EngineContextBase>

export const contextKey: InjectionKey<EngineContext> = Symbol.for('ddd:engine')

const must = <T>(v: T | undefined, error: string): T => {
  if(v==null){
    throw new Error(error)
  }
  return v
}

export function useContext(): EngineContext {
  return must(inject(contextKey), 'no context found')
}

export const parentViewIdKey: InjectionKey<string> = Symbol.for('ddd:view:parent:id')

export function useParentViewId(): string {
  return must(inject(parentViewIdKey), 'no parentViewId no found')
}

export function provideParentViewId(id: string | undefined): void {
  if (id) {
    provide(parentViewIdKey, id)
  }
}

const blockIdKey: InjectionKey<string> = Symbol.for('ddd:block:id')

export const useBlockId = (): string => {
  return must(inject(blockIdKey), "no block found")
}

export function provideBlockId(id: string): void {
  provide(blockIdKey, id)
}

type ReactingDomRect = InjectionKey<UnwrapNestedRefs<Partial<DOMRect>>>
export const currentViewDOMRectKey: ReactingDomRect = Symbol.for('ddd:view:current:domrect')
export const parentViewDOMRectKey: ReactingDomRect = Symbol.for('ddd:view:parent:domrect')

export function useCurrentViewDOMRect(): UnwrapNestedRefs<Partial<DOMRect>> {
  return must(inject(currentViewDOMRectKey), 'no DOMReact found')
}

export function useParentViewDOMRect(): UnwrapNestedRefs<Partial<DOMRect>> {
  return must(inject(parentViewDOMRectKey), 'no DOMReact found')
}

export function getModule(ctx: UnwrapNestedRefs<EngineContextBase>, mid: string): Module | undefined {
  for(const category of ctx.categories){
    for(const module of category.modules){
      if(module.mid === mid) {
        return module
      }
    }
  }
  return undefined
}

export function useModule(){
  const ctx = useContext()

  return computed((): Module | undefined => {
    for(const block of ctx.blocks){
      if(block.vid !== ctx.activeBlockId){
        continue
      }
      const module = getModule(ctx,block.mid)
      if (module != null) {
        return module
      }
    }
    return undefined
  })
}

export function useScale() {
  const ctx = useContext()

  const minScaleValue = 0.6
  const maxScaleValue = 3.0
  const scaleStep = 0.2
  const setScale = (delta: number): void => {
    ctx.canvas.scale = Math.max(
      Math.min(
        ctx.canvas.scale + scaleStep * delta,
        maxScaleValue,
      ),
      minScaleValue,
    )
  }

  return {
    min: minScaleValue,
    max: maxScaleValue,
    step: scaleStep,
    incr: () => setScale(1),
    decr: () => setScale(-1),
  }
}

export interface TreeData {
  key: string | number
  value: any
}

type RefTreeData = InjectionKey<UnwrapNestedRefs<TreeData>|WritableComputedRef<TreeData>|null>

const treeDataKey: RefTreeData = Symbol.for('ddd:view:tree:data')

export function provideTreeData(data: UnwrapNestedRefs<TreeData> | WritableComputedRef<TreeData> | null) {
  provide(treeDataKey, data)
}

export function useSource<T = unknown>(source: string | undefined, fallback?: T): WritableComputedRef<T | undefined> {
  const ctx = useContext()
  const treeData = inject(treeDataKey)
  const blockId = useBlockId()

  return computed<T | undefined>({

    get(): T | undefined {
      console.log('get',source,'source',blockId || ctx.activeBlockId,'id')
      if (!source) {
        return fallback
      }
      if (source === '$index' || source === '$key') {
        console.log(unref(treeData))
        return unref(treeData)?.key as T
      }
      if (source.startsWith('.')) {
        return valueOf(unref(treeData)?.value, source.slice(1)) ?? fallback
      }
      const id = blockId || ctx.activeBlockId
      if (!id) {
        return fallback
      }
      return valueOf(ctx.sources[id], source) ?? fallback
    },
    set(value: T | undefined) {
      console.log('set',value,'value',source,'source',blockId || ctx.activeBlockId,'id')
      if (!source || source.startsWith("$")) {
        return
      }
      if (source.startsWith('.')) {
        // if (treeData != null) {
        //   setValue(unref(treeData)?.value, source.slice(1), value)
        // }
        console.warn("不可以给 TreeData 赋值")
        return
      }
      const id = blockId || ctx.activeBlockId
      // console.log({id, source, value})
      if (id) {
        if (!ctx.sources[id]) {
          ctx.sources[id] = {}
        }
        setValue(ctx.sources[id], source, value)
      }
    }
  })
}

export function eachViewTree(view: ViewChildren | undefined, fn: (view: View) => void) {
  if (view == null || typeof view === 'string' || typeof view === 'number') {
    return
  }
  if (isRefer(view)) {
    // TODO ...
    return
  }
  if (Array.isArray(view)) {
    view.map(v => eachViewTree(v, fn))
    return
  }
  fn(view)
  eachViewTree(view.children, fn)
}
